from pdme.model.model import DipoleModel
from pdme.model.fixed_magnitude_model import SingleDipoleFixedMagnitudeModel
from pdme.model.multidipole_fixed_magnitude_model import (
	MultipleDipoleFixedMagnitudeModel,
)
from pdme.model.random_count_multidipole_fixed_magnitude_model import (
	RandomCountMultipleDipoleFixedMagnitudeModel,
)

from pdme.model.log_spaced_random_choice_model import (
	LogSpacedRandomCountMultipleDipoleFixedMagnitudeModel,
)

__all__ = [
	"DipoleModel",
	"SingleDipoleFixedMagnitudeModel",
	"MultipleDipoleFixedMagnitudeModel",
	"RandomCountMultipleDipoleFixedMagnitudeModel",
	"LogSpacedRandomCountMultipleDipoleFixedMagnitudeModel",
]



from dataclasses import dataclass
import numpy
import numpy.typing


@dataclass
class DotPairMeasurement:
	"""
	Representation of a dot measuring oscillating dipoles.

	Parameters
	----------
	v : float
	The voltage measured at the dot.

	r1 : numpy.ndarray
	The position of the first dot.

	r2 : numpy.ndarray
	The position of the second dot.

	f : float
	The measurement frequency.
	"""

	v: float
	r1: numpy.ndarray
	r2: numpy.ndarray
	f: float

	def __post_init__(self) -> None:
		self.r1 = numpy.array(self.r1)
		self.r2 = numpy.array(self.r2)


@dataclass
class DotPairRangeMeasurement:
	"""
	Representation of a dot measuring oscillating dipoles.

	Parameters
	----------
	v_low : float
	The lower range of voltage measured at the dot.

	v_high : float
	The upper range of voltage measured at the dot.

	r1 : numpy.ndarray
	The position of the first dot.

	r2 : numpy.ndarray
	The position of the second dot.

	f : float
	The measurement frequency.
	"""

	v_low: float
	v_high: float
	r1: numpy.ndarray
	r2: numpy.ndarray
	f: float

	def __post_init__(self) -> None:
		self.r1 = numpy.array(self.r1)
		self.r2 = numpy.array(self.r2)
		if self.v_low > self.v_high:
			self.v_low, self.v_high = self.v_high, self.v_low



import numpy
import numpy.typing
import itertools
from typing import Sequence, Tuple


def inputs_with_frequency_range(
	dots: Sequence[numpy.typing.ArrayLike], frequency_range: Sequence[float]
) -> Sequence[Tuple[numpy.typing.ArrayLike, float]]:
	return list(itertools.chain(*[[(dot, f) for f in frequency_range] for dot in dots]))


def input_pairs_with_frequency_range(
	dots: Sequence[numpy.typing.ArrayLike], frequency_range: Sequence[float]
) -> Sequence[Tuple[numpy.typing.ArrayLike, numpy.typing.ArrayLike, float]]:
	all_pairs = itertools.combinations(dots, 2)
	return list(
		itertools.chain(
			*[[(dot1, dot2, f) for f in frequency_range] for (dot1, dot2) in all_pairs]
		)
	)



import numpy
import numpy.random
from pdme.measurement import (
	OscillatingDipoleArrangement,
)
import logging


_logger = logging.getLogger(__name__)


class DipoleModel:
	"""
	Interface for models based on dipoles.
	Some concepts are kept specific for dipole-based models, even though other types of models could be useful later on.
	"""

	def get_dipoles(
		self, max_frequency: float, rng: numpy.random.Generator = None
	) -> OscillatingDipoleArrangement:
		"""
		For a particular maximum frequency, gets a dipole arrangement based on the model that uniformly distributes its choices according to the model.
		If no rng is passed in, uses some default, but you might not want that.
		Frequencies should be chosen uniformly on range of (0, max_frequency).
		"""
		raise NotImplementedError

	def get_monte_carlo_dipole_inputs(
		self, n: int, max_frequency: float, rng: numpy.random.Generator = None
	) -> numpy.ndarray:
		"""
		For a given DipoleModel, gets a set of dipole collections as a monte_carlo_n x dipole_count x 7 numpy array.
		"""
		raise NotImplementedError



import numpy
import numpy.random
from pdme.model.model import DipoleModel
from pdme.measurement import (
	OscillatingDipole,
	OscillatingDipoleArrangement,
)


class MultipleDipoleFixedMagnitudeModel(DipoleModel):
	"""
	Model of multiple oscillating dipoles with a fixed magnitude, but free rotation.

	Parameters
	----------
	pfixed : float
	The fixed dipole magnitude.

	n : int
	The number of dipoles.
	"""

	def __init__(
		self,
		xmin: float,
		xmax: float,
		ymin: float,
		ymax: float,
		zmin: float,
		zmax: float,
		pfixed: float,
		n: int,
	) -> None:
		self.xmin = xmin
		self.xmax = xmax
		self.ymin = ymin
		self.ymax = ymax
		self.zmin = zmin
		self.zmax = zmax
		self.pfixed = pfixed
		self.rng = numpy.random.default_rng()
		self.n = n

	def __repr__(self) -> str:
		return f"MultipleDipoleFixedMagnitudeModel({self.xmin}, {self.xmax}, {self.ymin}, {self.ymax}, {self.zmin}, {self.zmax}, {self.pfixed}, {self.n})"

	def get_dipoles(
		self, max_frequency: float, rng_to_use: numpy.random.Generator = None
	) -> OscillatingDipoleArrangement:
		rng: numpy.random.Generator
		if rng_to_use is None:
			rng = self.rng
		else:
			rng = rng_to_use

		dipoles = []

		for i in range(self.n):
			theta = numpy.arccos(rng.uniform(-1, 1))
			phi = rng.uniform(0, 2 * numpy.pi)
			px = self.pfixed * numpy.sin(theta) * numpy.cos(phi)
			py = self.pfixed * numpy.sin(theta) * numpy.sin(phi)
			pz = self.pfixed * numpy.cos(theta)
			s_pts = numpy.array(
				(
					rng.uniform(self.xmin, self.xmax),
					rng.uniform(self.ymin, self.ymax),
					rng.uniform(self.zmin, self.zmax),
				)
			)
			frequency = rng.uniform(0, max_frequency)

			dipoles.append(
				OscillatingDipole(numpy.array([px, py, pz]), s_pts, frequency)
			)
		return OscillatingDipoleArrangement(dipoles)

	def get_monte_carlo_dipole_inputs(
		self,
		monte_carlo_n: int,
		max_frequency: float,
		rng_to_use: numpy.random.Generator = None,
	) -> numpy.ndarray:

		rng: numpy.random.Generator
		if rng_to_use is None:
			rng = self.rng
		else:
			rng = rng_to_use

		shape = (monte_carlo_n, self.n)
		theta = 2 * numpy.pi * rng.random(shape)
		phi = numpy.arccos(2 * rng.random(shape) - 1)
		px = self.pfixed * numpy.cos(theta) * numpy.sin(phi)
		py = self.pfixed * numpy.sin(theta) * numpy.sin(phi)
		pz = self.pfixed * numpy.cos(phi)

		sx = rng.uniform(self.xmin, self.xmax, shape)
		sy = rng.uniform(self.ymin, self.ymax, shape)
		sz = rng.uniform(self.zmin, self.zmax, shape)

		w = rng.uniform(1, max_frequency, shape)

		return numpy.stack([px, py, pz, sx, sy, sz, w], axis=-1)



import numpy
import logging


_logger = logging.getLogger(__name__)


def fast_s_nonlocal(
	dot_pair_inputs: numpy.ndarray, dipoles: numpy.ndarray
) -> numpy.ndarray:
	"""
	No error correction here baby.
	"""
	ps = dipoles[:, 0:3]
	ss = dipoles[:, 3:6]
	ws = dipoles[:, 6]

	_logger.debug(f"ps: {ps}")
	_logger.debug(f"ss: {ss}")
	_logger.debug(f"ws: {ws}")

	r1s = dot_pair_inputs[:, 0, 0:3]
	r2s = dot_pair_inputs[:, 1, 0:3]
	f1s = dot_pair_inputs[:, 0, 3]
	f2s = dot_pair_inputs[:, 1, 3]

	if (f1s != f2s).all():
		raise ValueError(f"Dot pair frequencies are inconsistent: {dot_pair_inputs}")

	diffses1 = r1s - ss[:, None]
	diffses2 = r2s - ss[:, None]
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"diffses1: {diffses1}")
		_logger.debug(f"diffses2: {diffses2}")

	norms1 = numpy.linalg.norm(diffses1, axis=2) ** 3
	norms2 = numpy.linalg.norm(diffses2, axis=2) ** 3
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"norms1: {norms1}")
		_logger.debug(f"norms2: {norms2}")

	alphses1 = numpy.einsum("...ji, ...i", diffses1, ps) / norms1
	alphses2 = numpy.einsum("...ji, ...i", diffses2, ps) / norms2
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"alphses1: {alphses1}")
		_logger.debug(f"alphses2: {alphses2}")

	bses = (1 / numpy.pi) * (ws[:, None] / (f1s**2 + ws[:, None] ** 2))
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"bses: {bses}")

	return alphses1 * alphses2 * bses


def fast_s_nonlocal_dipoleses(
	dot_pair_inputs: numpy.ndarray, dipoleses: numpy.ndarray
) -> numpy.ndarray:
	"""
	No error correction here baby.
	"""
	ps = dipoleses[:, :, 0:3]
	ss = dipoleses[:, :, 3:6]
	ws = dipoleses[:, :, 6]

	_logger.debug(f"ps: {ps}")
	_logger.debug(f"ss: {ss}")
	_logger.debug(f"ws: {ws}")

	r1s = dot_pair_inputs[:, 0, 0:3]
	r2s = dot_pair_inputs[:, 1, 0:3]
	f1s = dot_pair_inputs[:, 0, 3]
	f2s = dot_pair_inputs[:, 1, 3]

	if (f1s != f2s).all():
		raise ValueError(f"Dot pair frequencies are inconsistent: {dot_pair_inputs}")

	diffses1 = r1s[:, None] - ss[:, None, :]
	diffses2 = r2s[:, None] - ss[:, None, :]
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"diffses1: {diffses1}")
		_logger.debug(f"diffses2: {diffses2}")

	norms1 = numpy.linalg.norm(diffses1, axis=3) ** 3
	norms2 = numpy.linalg.norm(diffses2, axis=3) ** 3
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"norms1: {norms1}")
		_logger.debug(f"norms2: {norms2}")

	alphses1 = numpy.einsum("abcd,acd->abc", diffses1, ps) / norms1
	alphses2 = numpy.einsum("abcd,acd->abc", diffses2, ps) / norms2
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"alphses1: {alphses1}")
		_logger.debug(f"alphses2: {alphses2}")

	bses = (1 / numpy.pi) * (ws[:, None, :] / (f1s[:, None] ** 2 + ws[:, None, :] ** 2))
	if _logger.isEnabledFor(logging.DEBUG):
		_logger.debug(f"bses: {bses}")

	_logger.debug(f"Raw pair calc: [{alphses1 * alphses2 * bses}]")
	return numpy.einsum("...j->...", alphses1 * alphses2 * bses)



import numpy
import logging


_logger = logging.getLogger(__name__)


def fast_vs_for_dipoles(
	dot_inputs: numpy.ndarray, dipoles: numpy.ndarray
) -> numpy.ndarray:
	"""
	No error correction here baby.
	Expects dot_inputs to be numpy array of [rx, ry, rz, f] entries, so a n by 4 where n is number of measurement points.
	"""
	ps = dipoles[:, 0:3]
	ss = dipoles[:, 3:6]
	ws = dipoles[:, 6]

	_logger.debug(f"ps: {ps}")
	_logger.debug(f"ss: {ss}")
	_logger.debug(f"ws: {ws}")

	rs = dot_inputs[:, 0:3]
	fs = dot_inputs[:, 3]

	diffses = rs - ss[:, None]

	_logger.debug(f"diffses: {diffses}")
	norms = numpy.linalg.norm(diffses, axis=2) ** 3
	_logger.debug(f"norms: {norms}")
	ases = (numpy.einsum("...ji, ...i", diffses, ps) / norms) ** 2
	_logger.debug(f"ases: {ases}")

	bses = (1 / numpy.pi) * (ws[:, None] / (fs**2 + ws[:, None] ** 2))
	_logger.debug(f"bses: {bses}")
	return ases * bses


def fast_vs_for_dipoleses(
	dot_inputs: numpy.ndarray, dipoleses: numpy.ndarray
) -> numpy.ndarray:
	"""
	No error correction here baby.
	Expects dot_inputs to be numpy array of [rx, ry, rz, f] entries, so a n by 4 where n is number of measurement points.

	Dipoleses are expected to be array of arrays of arrays: list of sets of dipoles which are part of a single arrangement to be added together.
	"""
	ps = dipoleses[:, :, 0:3]
	ss = dipoleses[:, :, 3:6]
	ws = dipoleses[:, :, 6]

	_logger.debug(f"ps: {ps}")
	_logger.debug(f"ss: {ss}")
	_logger.debug(f"ws: {ws}")

	rs = dot_inputs[:, 0:3]
	fs = dot_inputs[:, 3]

	diffses = rs[:, None] - ss[:, None, :]

	_logger.debug(f"diffses: {diffses}")
	norms = numpy.linalg.norm(diffses, axis=3) ** 3
	_logger.debug(f"norms: {norms}")
	ases = (numpy.einsum("abcd,acd->abc", diffses, ps) / norms) ** 2
	_logger.debug(f"ases: {ases}")

	bses = (1 / numpy.pi) * (ws[:, None, :] / (fs[:, None] ** 2 + ws[:, None, :] ** 2))
	_logger.debug(f"bses: {bses}")
	return numpy.einsum("...j->...", ases * bses)


def between(a: numpy.ndarray, low: numpy.ndarray, high: numpy.ndarray) -> numpy.ndarray:
	"""
	Intended specifically for the case where a is a list of arrays, and each array must be between the single array low and high, but without error checking.
	"""
	return numpy.all(numpy.logical_and(low < a, high > a), axis=1)



from pdme.measurement.dot_measure import DotMeasurement, DotRangeMeasurement
from pdme.measurement.dot_pair_measure import (
	DotPairMeasurement,
	DotPairRangeMeasurement,
)
from pdme.measurement.oscillating_dipole import (
	OscillatingDipole,
	OscillatingDipoleArrangement,
)
from pdme.measurement.input_types import DotInput, DotPairInput


__all__ = [
	"DotMeasurement",
	"DotRangeMeasurement",
	"DotPairMeasurement",
	"DotPairRangeMeasurement",
	"OscillatingDipole",
	"OscillatingDipoleArrangement",
	"DotInput",
	"DotPairInput",
]



import numpy
import numpy.random
from pdme.model.model import DipoleModel
from pdme.measurement import (
	OscillatingDipole,
	OscillatingDipoleArrangement,
)


class LogSpacedRandomCountMultipleDipoleFixedMagnitudeModel(DipoleModel):
	"""
	Model of multiple oscillating dipoles with a fixed magnitude, but free rotation. Spaced logarithmically.

	Parameters
	----------

	wexp_min: log-10 lower bound for dipole frequency
	wexp_min: log-10 upper bound for dipole frequency

	pfixed : float
	The fixed dipole magnitude.

	n_max : int
	The maximum number of dipoles.

	prob_occupancy : float
	The probability of dipole occupancy
	"""

	def __init__(
		self,
		xmin: float,
		xmax: float,
		ymin: float,
		ymax: float,
		zmin: float,
		zmax: float,
		wexp_min: float,
		wexp_max: float,
		pfixed: float,
		n_max: int,
		prob_occupancy: float,
	) -> None:
		self.xmin = xmin
		self.xmax = xmax
		self.ymin = ymin
		self.ymax = ymax
		self.zmin = zmin
		self.zmax = zmax
		self.wexp_min = wexp_min
		self.wexp_max = wexp_max
		self.pfixed = pfixed
		self.rng = numpy.random.default_rng()
		self.n_max = n_max
		if prob_occupancy >= 1 or prob_occupancy <= 0:
			raise ValueError(
				f"The probability of a dipole site occupancy must be between 0 and 1, got {prob_occupancy}"
			)
		self.prob_occupancy = prob_occupancy

	def __repr__(self) -> str:
		return f"LogSpacedRandomCountMultipleDipoleFixedMagnitudeModel({self.xmin}, {self.xmax}, {self.ymin}, {self.ymax}, {self.zmin}, {self.zmax}, {self.wexp_min}, {self.wexp_max}, {self.pfixed}, {self.n_max}, {self.prob_occupancy})"

	def get_dipoles(
		self, max_frequency: float, rng_to_use: numpy.random.Generator = None
	) -> OscillatingDipoleArrangement:
		rng: numpy.random.Generator
		if rng_to_use is None:
			rng = self.rng
		else:
			rng = rng_to_use

		dipoles = []

		n = rng.binomial(self.n_max, self.prob_occupancy)

		for i in range(n):
			theta = numpy.arccos(rng.uniform(-1, 1))
			phi = rng.uniform(0, 2 * numpy.pi)
			px = self.pfixed * numpy.sin(theta) * numpy.cos(phi)
			py = self.pfixed * numpy.sin(theta) * numpy.sin(phi)
			pz = self.pfixed * numpy.cos(theta)
			s_pts = numpy.array(
				(
					rng.uniform(self.xmin, self.xmax),
					rng.uniform(self.ymin, self.ymax),
					rng.uniform(self.zmin, self.zmax),
				)
			)
			frequency = 10 ** rng.uniform(self.wexp_min, self.wexp_max)

			dipoles.append(
				OscillatingDipole(numpy.array([px, py, pz]), s_pts, frequency)
			)
		return OscillatingDipoleArrangement(dipoles)

	def get_monte_carlo_dipole_inputs(
		self,
		monte_carlo_n: int,
		_: float,
		rng_to_use: numpy.random.Generator = None,
	) -> numpy.ndarray:

		rng: numpy.random.Generator
		if rng_to_use is None:
			rng = self.rng
		else:
			rng = rng_to_use

		shape = (monte_carlo_n, self.n_max)
		theta = 2 * numpy.pi * rng.random(shape)
		phi = numpy.arccos(2 * rng.random(shape) - 1)

		p_mask = rng.binomial(1, 0.5, shape)
		p_magnitude = self.pfixed * p_mask

		px = p_magnitude * numpy.cos(theta) * numpy.sin(phi)
		py = p_magnitude * numpy.sin(theta) * numpy.sin(phi)
		pz = p_magnitude * numpy.cos(phi)

		sx = rng.uniform(self.xmin, self.xmax, shape)
		sy = rng.uniform(self.ymin, self.ymax, shape)
		sz = rng.uniform(self.zmin, self.zmax, shape)

		w = 10 ** rng.uniform(self.wexp_min, self.wexp_max, shape)

		return numpy.stack([px, py, pz, sx, sy, sz, w], axis=-1)



from dataclasses import dataclass
import numpy
import numpy.typing
from typing import Sequence, List
from pdme.measurement.dot_measure import DotMeasurement, DotRangeMeasurement
from pdme.measurement.dot_pair_measure import (
	DotPairMeasurement,
	DotPairRangeMeasurement,
)
from pdme.measurement.input_types import DotInput, DotPairInput


@dataclass
class OscillatingDipole:
	"""
	Representation of an oscillating dipole, either known or guessed.

	Parameters
	----------
	p : numpy.ndarray
	The oscillating dipole moment, with overall sign arbitrary.

	s : numpy.ndarray
	The position of the dipole.

	w : float
	The oscillation frequency.
	"""

	p: numpy.ndarray
	s: numpy.ndarray
	w: float

	def __post_init__(self) -> None:
		"""
		Coerce the inputs into numpy arrays.
		"""
		self.p = numpy.array(self.p)
		self.s = numpy.array(self.s)

	def s_at_position(self, r: numpy.ndarray, f: float) -> float:
		"""
		Returns the noise potential at a point r, at some frequency f.

		Parameters
		----------
		r : numpy.ndarray
		The position of the dot.

		f : float
		The dot frequency to sample.
		"""
		return (self._alpha(r)) ** 2 * self._b(f)

	def _alpha(self, r: numpy.ndarray) -> float:
		diff = r - self.s
		return self.p.dot(diff) / (numpy.linalg.norm(diff) ** 3)

	def _b(self, f: float) -> float:
		return (1 / numpy.pi) * (self.w / (f**2 + self.w**2))

	def s_for_dot_pair(self, r1: numpy.ndarray, r2: numpy.ndarray, f: float) -> float:
		return self._alpha(r1) * self._alpha(r2) * self._b(f)

	def to_flat_array(self) -> numpy.ndarray:
		return numpy.concatenate([self.p, self.s, numpy.array([self.w])])


class OscillatingDipoleArrangement:
	"""
	A collection of oscillating dipoles, which we are interested in being able to characterise.

	Parameters
	--------
	dipoles : Sequence[OscillatingDipole]
	"""

	def __init__(self, dipoles: Sequence[OscillatingDipole]):
		self.dipoles = dipoles

	def get_dot_measurement(self, dot_input: DotInput) -> DotMeasurement:
		r = numpy.array(dot_input[0])
		f = dot_input[1]
		return DotMeasurement(
			sum([dipole.s_at_position(r, f) for dipole in self.dipoles]), r, f
		)

	def get_dot_pair_measurement(
		self, dot_pair_input: DotPairInput
	) -> DotPairMeasurement:
		r1 = numpy.array(dot_pair_input[0])
		r2 = numpy.array(dot_pair_input[1])
		f = dot_pair_input[2]
		return DotPairMeasurement(
			sum([dipole.s_for_dot_pair(r1, r2, f) for dipole in self.dipoles]),
			r1,
			r2,
			f,
		)

	def get_dot_measurements(
		self, dot_inputs: Sequence[DotInput]
	) -> List[DotMeasurement]:
		"""
		For a series of points, each with three coordinates and a frequency, return a list of the corresponding DotMeasurements.
		"""
		return [self.get_dot_measurement(dot_input) for dot_input in dot_inputs]

	def get_dot_pair_measurements(
		self, dot_pair_inputs: Sequence[DotPairInput]
	) -> List[DotPairMeasurement]:
		"""
		For a series of pairs of points, each with three coordinates and a frequency, return a list of the corresponding DotPairMeasurements.
		"""
		return [
			self.get_dot_pair_measurement(dot_pair_input)
			for dot_pair_input in dot_pair_inputs
		]

	def get_percent_range_dot_measurement(
		self, dot_input: DotInput, low_percent: float, high_percent: float
	) -> DotRangeMeasurement:
		r = numpy.array(dot_input[0])
		f = dot_input[1]
		return DotRangeMeasurement(
			low_percent * sum([dipole.s_at_position(r, f) for dipole in self.dipoles]),
			high_percent * sum([dipole.s_at_position(r, f) for dipole in self.dipoles]),
			r,
			f,
		)

	def get_percent_range_dot_measurements(
		self, dot_inputs: Sequence[DotInput], low_percent: float, high_percent: float
	) -> List[DotRangeMeasurement]:
		"""
		For a series of pairs of points, each with three coordinates and a frequency, and also a lower error range and upper error range, return a list of the corresponding DotPairRangeMeasurements.
		"""
		return [
			self.get_percent_range_dot_measurement(dot_input, low_percent, high_percent)
			for dot_input in dot_inputs
		]

	def get_percent_range_dot_pair_measurement(
		self, pair_input: DotPairInput, low_percent: float, high_percent: float
	) -> DotPairRangeMeasurement:
		r1 = numpy.array(pair_input[0])
		r2 = numpy.array(pair_input[1])
		f = pair_input[2]
		return DotPairRangeMeasurement(
			low_percent
			* sum([dipole.s_for_dot_pair(r1, r2, f) for dipole in self.dipoles]),
			high_percent
			* sum([dipole.s_for_dot_pair(r1, r2, f) for dipole in self.dipoles]),
			r1,
			r2,
			f,
		)

	def get_percent_range_dot_pair_measurements(
		self,
		pair_inputs: Sequence[DotPairInput],
		low_percent: float,
		high_percent: float,
	) -> List[DotPairRangeMeasurement]:
		"""
		For a series of pairs of points, each with three coordinates and a frequency, and also a lower error range and upper error range, return a list of the corresponding DotPairRangeMeasurements.
		"""
		return [
			self.get_percent_range_dot_pair_measurement(
				pair_input, low_percent, high_percent
			)
			for pair_input in pair_inputs
		]

	def to_numpy_array(self) -> numpy.ndarray:
		"""
		Returns a numpy array with the canonical representation of each dipole in a nx7 numpy array.
		"""
		return numpy.array([dipole.to_flat_array() for dipole in self.dipoles])
